import os.path
import secrets
import tempfile
from unittest.mock import MagicMock

import pytest
from pathlib import Path

from dagshub.streaming import DagsHubFilesystem
from dagshub.streaming.dataclasses import DagshubPath
from dagshub.streaming.errors import FilesystemAlreadyMountedError


@pytest.mark.parametrize(
    "path,expected",
    [
        ("regular/path/in/repo", False),
        (".", False),
        (".git/", True),
        (".git/and/then/some", True),
        (".dvc/", True),
        (".dvc/and/then/some", True),
        ("repo/file.dvc/", False),
        ("repo/file.git/", False),
        ("venv/lib/site-packages/some-package", True),
    ],
)
def test_passthrough_path(path, expected):
    fs_mock = MagicMock()
    path = DagshubPath(fs_mock, Path(os.path.abspath(path)), Path(path))
    actual = path.is_passthrough_path
    assert actual == expected


@pytest.mark.parametrize("a_path, b_path, create_folder", [
    (".", ".", False),
    (".", "./subpath", True),
    (".", "../", False)
], ids=["Same dir", "Sub dir", "Parent dir"])
def test_cant_mount_multiples(mock_api, a_path, b_path, create_folder):
    new_branch = "new"
    sha = secrets.token_hex(nbytes=20)
    mock_api.add_branch(new_branch, sha)
    _ = DagsHubFilesystem(project_root=a_path, repo_url="https://dagshub.com/user/repo")
    if create_folder:
        os.makedirs(b_path, exist_ok=True)
    with pytest.raises(FilesystemAlreadyMountedError):
        _ = DagsHubFilesystem(project_root=b_path, repo_url="https://dagshub.com/user/repo", branch=new_branch)


def test_can_mount_multiple_in_different_dirs(mock_api):
    new_branch = "new"
    sha = secrets.token_hex(nbytes=20)
    mock_api.add_branch(new_branch, sha)

    _ = DagsHubFilesystem(project_root=".", repo_url="https://dagshub.com/user/repo")

    other_path = tempfile.mkdtemp()
    resp = mock_api._default_endpoints_and_responses()[1]["list_root"]
    mock_api.get(url=f"/api/v1/repos/user/repo/content/{sha}/").mock(resp)

    _ = DagsHubFilesystem(project_root=other_path, repo_url="https://dagshub.com/user/repo", branch=new_branch)



import os

import pytest


def test_listdir(mock_api, repo_with_hooks):
    files = [("a.txt", "file"), ("b.txt", "file"), ("dir1", "dir")]
    path = "testdir"
    mock_api.add_dir(path, files)
    actual = os.listdir(path)
    expected = [f[0] for f in files]
    # Compare sets to ignore order of the lists
    assert set(actual) == set(expected)


def test_listdir_includes_local_files(mock_api, repo_with_hooks):
    path = "testdir"
    os.mkdir(path)
    with pytest.MonkeyPatch.context() as mp:
        mp.chdir(path)
        files = [("a.txt", "file"), ("b.txt", "file"), ("dir1", "dir")]
        mock_api.add_dir(path, files)
        add_file = "c.txt"
        with open(add_file, "w") as f:
            f.write("test")

        expected = [f[0] for f in files]
        expected.append(add_file)

        actual = os.listdir(".")
        assert set(expected) == set(actual)


def test_has_storage_bucket_paths(mock_api, repo_with_hooks):
    bucket_path = mock_api.storage_bucket_path
    bucket_path_elems = bucket_path.split("/")
    assert len(bucket_path_elems) == 2
    for root, dirs, files in os.walk("."):
        if root == ".":
            assert ".dagshub" in dirs
        elif root == "./.dagshub":
            assert "storage" in dirs
        elif root == "./.dagshub/storage":
            assert "s3" in dirs
        elif root == "./.dagshub/storage/s3":
            assert bucket_path_elems[0] in dirs
        elif root == f"./.dagshub/storage/s3/{bucket_path_elems[0]}":
            assert bucket_path_elems[1] in dirs


def test_lists_bucket_folder(mock_api, repo_with_hooks):
    files = [("a.txt", "file"), ("b.txt", "file"), ("dir1", "dir")]
    path = "testdir"
    mock_api.add_storage_dir(path, files)
    actual = os.listdir(f".dagshub/storage/s3/{mock_api.storage_bucket_path}/{path}")
    expected = [f[0] for f in files]
    assert set(actual) == set(expected)


def test_storage_pagination(mock_api, repo_with_hooks):
    files1 = [("a.txt", "file"), ("b.txt", "file"), ("dir1", "dir")]
    files2 = [("c.txt", "file")]
    path = "testdir"
    # Order is important, otherwise respx loops and returns the first result indefinitely
    mock_api.add_storage_dir(path, files2, from_token="aaa")
    mock_api.add_storage_dir(path, files1, next_token="aaa")

    expected = [f[0] for f in files1 + files2]

    actual = os.listdir(f".dagshub/storage/s3/{mock_api.storage_bucket_path}/{path}")
    assert set(actual) == set(expected)


def test_binary(mock_api, repo_with_hooks):
    files = [("a.txt", "file"), ("b.txt", "file"), ("dir1", "dir")]
    path = "testdir"
    mock_api.add_dir(path, files)
    actual = os.listdir(path.encode("utf-8"))
    expected = [f[0].encode("utf-8") for f in files]
    assert set(actual) == set(expected)
